@@ -1,11 +1,11 @@
 ;;; nov.el --- Featureful EPUB reader mode
 
-;; Copyright (C) 2017-2019 Vasilij Schneidermann <mail@vasilij.de>
+;; Copyright (C) 2017 Vasilij Schneidermann <mail@vasilij.de>
 
 ;; Author: Vasilij Schneidermann <mail@vasilij.de>
-;; URL: https://github.com/wasamasa/nov.el
-;; Version: 0.3.0
-;; Package-Requires: ((dash "2.12.0") (esxml "0.3.3") (emacs "24.4"))
+;; URL: https://depp.brause.cc/nov.el
+;; Version: 0.3.4
+;; Package-Requires: ((dash "2.12.0") (esxml "0.3.6") (emacs "24.4"))
 ;; Keywords: hypermedia, multimedia, epub
 
 ;; This file is NOT part of GNU Emacs.
@@ -46,11 +46,15 @@
 (require 'cl-lib)
 (require 'dash)
 (require 'esxml-query)
+(require 'image)
 (require 'shr)
 (require 'url-parse)
-(require 'recentf)
+(require 'xml)
+
 (require 'bookmark)
+(require 'imenu)
 (require 'org)
+(require 'recentf)
 
 (when (not (fboundp 'libxml-parse-xml-region))
   (message "Your Emacs wasn't compiled with libxml support"))
@@ -180,8 +184,8 @@ Each element of the stack is a list (NODEINDEX BUFFERPOS).")
 (defun nov-unzip-epub (directory filename)
   "Extract FILENAME into DIRECTORY.
 Unnecessary nesting is removed with `nov-unnest-directory'."
-  (let ((status (call-process nov-unzip-program nil nil t
-                              "-d" directory filename))
+  (let ((status (call-process nov-unzip-program nil "*nov unzip*" t
+                              "-od" directory filename))
         child)
     (while (setq child (nov-contains-nested-directory-p directory))
       (nov-unnest-directory directory child))
@@ -262,7 +266,7 @@ the specific type of unique identifier."
   "Return the the unique identifier for CONTENT."
   (let* ((name (nov-content-unique-identifier-name content))
          (selector (format "package>metadata>identifier[id='%s']"
-                           (regexp-quote name)))
+                           (esxml-query-css-escape name)))
          (id (car (esxml-node-children (esxml-query selector content)))))
     (when (not id)
       (error "Unique identifier not found by its name: %s" name))
@@ -347,14 +351,14 @@ Each alist item consists of the identifier and full path."
         (nov--content-epub2-files content manifest files)
       (nov--content-epub3-files content manifest files))))
 
-(defun nov--walk-ncx-node (node depth)
+(defun nov--walk-ncx-node (node)
   (let ((tag (esxml-node-tag node))
         (children (--filter (eq (esxml-node-tag it) 'navPoint)
                             (esxml-node-children node))))
     (cond
      ((eq tag 'navMap)
       (insert "<ol>\n")
-      (mapc (lambda (node) (nov--walk-ncx-node node (1+ depth))) children)
+      (mapc (lambda (node) (nov--walk-ncx-node node)) children)
       (insert "</ol>\n"))
      ((eq tag 'navPoint)
       (let* ((label-node (esxml-query "navLabel>text" node))
@@ -363,11 +367,13 @@ Each alist item consists of the identifier and full path."
              (label (car (esxml-node-children label-node))))
         (when (not href)
           (error "Navigation point is missing href attribute"))
-        (let ((link (format "<a href=\"%s\">%s</a>" href (or label href))))
+        (let ((link (format "<a href=\"%s\">%s</a>"
+                            (xml-escape-string href)
+                            (xml-escape-string (or label href)))))
           (if children
               (progn
                 (insert (format "<li>\n%s\n<ol>\n" link))
-                (mapc (lambda (node) (nov--walk-ncx-node node (1+ depth)))
+                (mapc (lambda (node) (nov--walk-ncx-node node))
                       children)
                 (insert (format "</ol>\n</li>\n")))
             (insert (format "<li>\n%s\n</li>\n" link)))))))))
@@ -376,7 +382,7 @@ Each alist item consists of the identifier and full path."
   "Convert NCX document at PATH to HTML."
   (let ((root (esxml-query "navMap" (nov-slurp path t))))
     (with-temp-buffer
-      (nov--walk-ncx-node root 0)
+      (nov--walk-ncx-node root)
       (buffer-string))))
 
 
@@ -437,32 +443,30 @@ Each alist item consists of the identifier and full path."
   (setq url (url-generic-parse-url url))
   (mapcar 'nov-urldecode (list (url-filename url) (url-target url))))
 
+;; adapted from `shr-rescale-image'
 (defun nov-insert-image (path alt)
   "Insert an image for PATH at point, falling back to ALT.
 This function honors `shr-max-image-proportion' if possible."
-  (cond
-   ((not (display-graphic-p))
-    (insert alt))
-   ;; TODO: add native resizing support once it's official
-   ((fboundp 'imagemagick-types)
-    ;; adapted from `shr-rescale-image'
-    (let ((edges (window-inside-pixel-edges
-                  (get-buffer-window (current-buffer)))))
-      (insert-image
-       (create-image path 'imagemagick nil
-                     :ascent 100
-                     :max-width (truncate (* shr-max-image-proportion
-                                             (- (nth 2 edges)
-                                                (nth 0 edges))))
-                     :max-height (truncate (* shr-max-image-proportion
-                                              (- (nth 3 edges)
-                                                 (nth 1 edges))))))))
-   (t
-    ;; `create-image' errors out for unsupported image types
-    (let ((image (ignore-errors (create-image path nil nil :ascent 100))))
-      (if image
-          (insert-image image)
-        (insert alt))))))
+  (let ((type (if (or (and (fboundp 'image-transforms-p) (image-transforms-p))
+                      (not (fboundp 'imagemagick-types)))
+                  nil
+                'imagemagick)))
+    (if (not (display-graphic-p))
+        (insert alt)
+      (-let* (((x1 y1 x2 y2) (window-inside-pixel-edges
+                              (get-buffer-window (current-buffer))))
+              (image
+               ;; `create-image' errors out for unsupported image types
+               (ignore-errors
+                 (create-image path type nil
+                               :ascent 100
+                               :max-width (truncate (* shr-max-image-proportion
+                                                       (- x2 x1)))
+                               :max-height (truncate (* shr-max-image-proportion
+                                                        (- y2 y1)))))))
+        (if image
+            (insert-image image)
+          (insert alt))))))
 
 (defvar nov-original-shr-tag-img-function
   (symbol-function 'shr-tag-img))
@@ -524,15 +528,13 @@ If the document path refers to an image (as determined by
 `image-type-file-name-regexps'), an image is inserted, otherwise
 the HTML is rendered with `nov-render-html-function'."
   (interactive)
-  (let* ((document (aref nov-documents nov-documents-index))
-         (id (car document))
-         (path (cdr document))
-         ;; HACK: this should be looked up in the manifest
-         (imagep (--find (string-match-p (car it) path)
-                         image-type-file-name-regexps))
-         ;; NOTE: allows resolving image references correctly
-         (default-directory (file-name-directory path))
-         buffer-read-only)
+  (-let* (((id . path) (aref nov-documents nov-documents-index))
+          ;; HACK: this should be looked up in the manifest
+          (imagep (--find (string-match-p (car it) path)
+                          image-type-file-name-regexps))
+          ;; NOTE: allows resolving image references correctly
+          (default-directory (file-name-directory path))
+          (buffer-read-only nil))
     (erase-buffer)
 
     (cond
@@ -648,23 +650,29 @@ the HTML is rendered with `nov-render-html-function'."
 
 (defun nov-visit-relative-file (filename target)
   "Visit the document as specified by FILENAME and TARGET."
-  (when (not (zerop (length filename)))
-    (let* ((current-path (cdr (aref nov-documents nov-documents-index)))
-           (directory (file-name-directory current-path))
-           (path (file-truename (nov-make-path directory filename)))
-           (index (nov-find-document
-                   (lambda (doc) (equal path (file-truename (cdr doc)))))))
-      (when (not index)
-        (error "Couldn't locate document"))
-      (setq nov-documents-index index)))
-  (let ((shr-target-id target))
-    (nov-goto-document nov-documents-index))
+  (let (index)
+    (when (not (zerop (length filename)))
+      (let* ((current-path (cdr (aref nov-documents nov-documents-index)))
+             (directory (file-name-directory current-path))
+             (path (file-truename (nov-make-path directory filename)))
+             (match (nov-find-document
+                     (lambda (doc) (equal path (file-truename (cdr doc)))))))
+        (when (not match)
+          (error "Couldn't locate document"))
+        (setq index match)))
+    (let ((shr-target-id target)) ; shr.el API for Emacs 27.1 and older
+      (nov-goto-document (or index nov-documents-index))))
   (when target
-    (let ((pos (next-single-property-change (point-min) 'shr-target-id)))
-      (when (not pos)
-        (error "Couldn't locate target"))
-      (goto-char pos)
-      (recenter (1- (max 1 scroll-margin))))))
+    (let ((pos (point-min))
+          done)
+      (while (and (not done)
+                  (setq pos (next-single-property-change pos 'shr-target-id)))
+        (when (equal (get-text-property pos 'shr-target-id) target)
+          (goto-char pos)
+          (recenter (1- (max 1 scroll-margin)))
+          (setq done t)))
+      (when (not done)
+        (error "Couldn't locate target")))))
 
 ;; adapted from `shr-browse-url'
 (defun nov-browse-url (&optional mouse-event)
@@ -722,28 +730,28 @@ Saving is only done if `nov-save-place-file' is set."
   (interactive)
   (or nov-history
       (user-error "This is the first document you looked at"))
-  (let ((history-forward
-         (cons (list nov-documents-index (point))
-               nov-history-forward))
-        (index (car (car nov-history)))
-        (opoint (cadr (car nov-history))))
+  (-let ((history-forward
+          (cons (list nov-documents-index (point))
+                nov-history-forward))
+         ((index opoint) (car nov-history)))
     (setq nov-history (cdr nov-history))
     (nov-goto-document index)
     (setq nov-history (cdr nov-history))
     (setq nov-history-forward history-forward)
-    (goto-char opoint)))
+    (goto-char opoint)
+    (recenter (1- (max 1 scroll-margin)))))
 
 (defun nov-history-forward ()
   "Go forward in the history of visited documents."
   (interactive)
   (or nov-history-forward
       (user-error "This is the last document you looked at"))
-  (let ((history-forward (cdr nov-history-forward))
-        (index (car (car nov-history-forward)))
-        (opoint (cadr (car nov-history-forward))))
+  (-let ((history-forward (cdr nov-history-forward))
+         ((index opoint) (car nov-history-forward)))
     (nov-goto-document index)
     (setq nov-history-forward history-forward)
-    (goto-char opoint)))
+    (goto-char opoint)
+    (recenter (1- (max 1 scroll-margin)))))
 
 ;;;###autoload
 (define-derived-mode nov-mode special-mode "EPUB"
@@ -758,9 +766,10 @@ Saving is only done if `nov-save-place-file' is set."
     (when (not (integerp exit-code))
       (nov-clean-up)
       (error "EPUB extraction aborted by signal %s" exit-code))
-    (when (not (zerop exit-code))
+    (when (> exit-code 1) ; exit code 1 is most likely a warning
       (nov-clean-up)
-      (error "EPUB extraction failed with exit code %d" exit-code)))
+      (error "EPUB extraction failed with exit code %d (see *nov unzip* buffer)"
+             exit-code)))
   (when (not (nov-epub-valid-p nov-temp-dir))
     (nov-clean-up)
     (error "Invalid EPUB file"))
@@ -797,6 +806,7 @@ Saving is only done if `nov-save-place-file' is set."
 ;;; recentf interop
 
 (defun nov-add-to-recentf ()
+  "Add real path to recentf list if possible."
   (when nov-file-name
     (recentf-add-file nov-file-name)))
 
@@ -816,7 +826,7 @@ Saving is only done if `nov-save-place-file' is set."
   (nov-render-document)
   (goto-char point))
 
-;; Bookmark Integration
+;; Bookmark interop
 (defun nov-bookmark-make-record  ()
   "Create a bookmark epub record."
   (cons (buffer-name)
@@ -825,6 +835,7 @@ Saving is only done if `nov-save-place-file' is set."
           (position . ,(point))
           (handler . nov-bookmark-jump-handler))))
 
+;;;###autoload
 (defun nov-bookmark-jump-handler (bmk)
   "The bookmark handler-function interface for bookmark BMK.
 
@@ -835,9 +846,10 @@ See also `nov-bookmark-make-record'."
     (nov--find-file file index position)))
 
 
-;;; org interop
+;;; Org interop
 
 (defun nov-org-link-follow (path)
+  "Follow nov: link designated by PATH."
   (if (string-match "^\\(.*\\)::\\([0-9]+\\):\\([0-9]+\\)$" path)
       (let ((file (match-string 1 path))
             (index (string-to-number (match-string 2 path)))
@@ -846,18 +858,63 @@ See also `nov-bookmark-make-record'."
     (error "Invalid nov.el link")))
 
 (defun nov-org-link-store ()
+  "Store current EPUB location as nov: link."
   (when (and (eq major-mode 'nov-mode) nov-file-name)
     (when (not (integerp nov-documents-index))
       (setq nov-documents-index 0))
-    (org-store-link-props
-     :type "nov"
-     :link (format "nov:%s::%d:%d" nov-file-name nov-documents-index (point))
-     :description (format "EPUB file at %s" nov-file-name))))
+    (let ((org-store-props-function
+           (if (fboundp 'org-link-store-props)
+               'org-link-store-props
+             'org-store-link-props))
+          (link (format "nov:%s::%d:%d"
+                        nov-file-name
+                        nov-documents-index
+                        (point)))
+          (description (format "EPUB file at %s" nov-file-name)))
+      (funcall org-store-props-function
+               :type "nov"
+               :link link
+               :description description))))
 
-(org-link-set-parameters
- "nov"
- :follow 'nov-org-link-follow
- :store 'nov-org-link-store)
+(cond
+ ((fboundp 'org-link-set-parameters)
+  (org-link-set-parameters
+   "nov"
+   :follow 'nov-org-link-follow
+   :store 'nov-org-link-store))
+ ((fboundp 'org-add-link-type)
+  (org-add-link-type "nov" 'nov-org-link-follow)
+  (add-hook 'org-store-link-functions 'nov-org-link-store)))
+
+
+;;; Imenu interop
+
+(defun nov-imenu-goto-function (_name filename target)
+  "Visit imenu item using FILENAME and TARGET."
+  (nov-visit-relative-file filename target))
+
+(defun nov-imenu-create-index ()
+  "Generate Imenu index."
+  (require 'esxml)
+  (let* ((toc-path (cdr (aref nov-documents 0)))
+         (ncxp (version< nov-epub-version "3.0"))
+         (toc (with-temp-buffer
+                (if ncxp
+                    (insert (nov-ncx-to-html toc-path))
+                  (insert-file-contents toc-path))
+                (libxml-parse-html-region (point-min) (point-max)))))
+    (mapcar
+     (lambda (node)
+       (-let* ((href (esxml-node-attribute 'href node))
+               (label (mapconcat 'string-trim-whitespace
+                                 (esxml-find-descendants #'stringp node) " "))
+               ((filename target) (nov-url-filename-and-target href)))
+         (list label filename 'nov-imenu-goto-function target)))
+     (esxml-query-all "a" toc))))
+
+(defun nov-imenu-setup ()
+  (setq imenu-create-index-function 'nov-imenu-create-index))
+(add-hook 'nov-mode-hook 'nov-imenu-setup)
 
 (provide 'nov)
 ;;; nov.el ends here
